import filecmp
import os

from numpy.testing import assert_equal

from iseq.example import example_filepath
from iseq.gff import read as read_gff


def test_gff_read():
    items = read_gff(example_filepath("duplicate.gff")).items()
    assert_equal(len(items), 14)
    assert_equal(items[3].seqid, "GALNBKIG_00914_ont_01_plus_strand")
    assert_equal(items[6].end, 474)


def test_gff_deduplicate(tmp_path):
    os.chdir(tmp_path)

    gff = read_gff(example_filepath("duplicate.gff"))
    gff.deduplicate()

    gff.write_file("output.gff")

    dedup_file = example_filepath("deduplicate.gff")
    assert_equal(filecmp.cmp(dedup_file, "output.gff", shallow=False), True)



from iseq.example import example_filepath
from iseq.tblout import TBLData


def test_tblout():
    with open(example_filepath("tblout.txt"), "r") as file:
        reader = TBLData(file)

        row = reader.find("item2", query_name="Octapeptide", query_acc="PF03373.14")
        assert row.target_name == "item2"
        assert row.full_sequence.e_value == "1.2e-07"
        assert row.best_1_domain.e_value == "1.2e-07"

        row = reader.find("item3", query_name="Octapeptide", query_acc="PF03373.14")
        assert row.target_name == "item3"
        assert row.full_sequence.e_value == "1.2e-07"
        assert row.best_1_domain.e_value == "1.2e-07"



import os
import re
from collections import OrderedDict
from pathlib import Path
from typing import IO

import click
from fasta_reader import FASTAWriter, open_fasta
from hmmer_reader import num_models, open_hmmer
from nmm import AminoAlphabet, BaseAlphabet, CanonicalAminoAlphabet, GeneticCode
from tqdm import tqdm

from iseq.alphabet import infer_fasta_alphabet, infer_hmmer_alphabet
from iseq.hmmer_model import HMMERModel
from iseq.hmmsearch import HMMSearch
from iseq.protein import create_profile
from iseq.tblout import TBLData

from .debug_writer import DebugWriter
from .output_writer import OutputWriter


@click.command()
@click.argument(
    "profile",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
@click.argument("target", type=click.File("r"))
@click.option(
    "--epsilon", type=float, default=1e-2, help="Indel probability. Defaults to 1e-2."
)
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save results to OUTPUT (GFF format).",
    default="output.gff",
)
@click.option(
    "--ocodon",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save codon sequences to OCODON (FASTA format).",
    default="ocodon.fasta",
)
@click.option(
    "--oamino",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save amino acid sequences to OAMINO (FASTA format).",
    default="oamino.fasta",
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable standard output.", default=False,
)
@click.option(
    "--window",
    type=int,
    help="Window length. Defaults to zero, which means no window.",
    default=0,
)
@click.option(
    "--odebug",
    type=click.File("w"),
    help="Save debug info into a tab-separated values file.",
    default=os.devnull,
)
@click.option(
    "--e-value/--no-e-value",
    help="Enable E-value computation. Defaults to True.",
    default=True,
)
def pscan(
    profile,
    target,
    epsilon: float,
    output,
    ocodon,
    oamino,
    quiet,
    window: int,
    odebug,
    e_value: bool,
):
    """
    Search nucleotide sequence(s) against a protein profiles database.

    An OUTPUT line determines an association between a TARGET subsequence and
    a PROFILE protein profile. An association maps a target subsequence to a
    profile and represents a potential homology. Expect many false positive
    associations as we are not filtering out by statistical significance.
    """

    owriter = OutputWriter(output)
    cwriter = FASTAWriter(ocodon)
    awriter = FASTAWriter(oamino)
    dwriter = DebugWriter(odebug)

    with open(profile, "r") as file:
        profile_abc = _infer_profile_alphabet(file)
    target_abc = infer_target_alphabet(target)

    assert isinstance(target_abc, BaseAlphabet) and isinstance(
        profile_abc, AminoAlphabet
    )

    gcode = GeneticCode(target_abc, CanonicalAminoAlphabet())

    with open_fasta(target) as fasta:
        targets = list(fasta)

    total = num_models(profile)
    for plain_model in tqdm(open_hmmer(profile), total=total, disable=quiet):
        model = HMMERModel(plain_model)
        prof = create_profile(model, gcode.base_alphabet, window, epsilon)

        for tgt in targets:
            seq = prof.create_sequence(tgt.sequence.encode())
            search_results = prof.search(seq)
            ifragments = search_results.ifragments()
            seqid = f"{tgt.defline.split()[0]}"

            for ifrag in ifragments:
                start = ifrag.interval.start
                stop = ifrag.interval.stop
                item_id = owriter.write_item(
                    seqid,
                    prof.profid,
                    start,
                    stop,
                    prof.window_length,
                    {"Epsilon": epsilon},
                )
                codon_frag = ifrag.fragment.decode()
                cwriter.write_item(item_id, str(codon_frag.sequence))
                amino_frag = codon_frag.decode(gcode)
                awriter.write_item(item_id, str(amino_frag.sequence))

            if odebug is not os.devnull:
                for i in search_results.debug_table():
                    dwriter.write_row(seqid, i)

    owriter.close()
    cwriter.close()
    awriter.close()
    odebug.close_intelligently()

    if e_value:
        hmmsearch = HMMSearch()
        tbldata = hmmsearch.search(Path(profile), Path(oamino))
        update_gff_file(output, tbldata)


def _infer_profile_alphabet(profile: IO[str]):
    hmmer = open_hmmer(profile)
    hmmer_alphabet = infer_hmmer_alphabet(hmmer)
    profile.seek(0)
    if hmmer_alphabet is None:
        raise click.UsageError("Could not infer alphabet from PROFILE.")
    return hmmer_alphabet


def infer_target_alphabet(target: IO[str]):
    fasta = open_fasta(target)
    target_alphabet = infer_fasta_alphabet(fasta)
    target.seek(0)
    if target_alphabet is None:
        raise click.UsageError("Could not infer alphabet from TARGET.")
    return target_alphabet


def update_gff_file(filepath, tbldata: TBLData):
    import in_place

    with in_place.InPlace(filepath) as file:
        for row in file:
            row = row.rstrip()
            if row.startswith("#"):
                file.write(row)
                file.write("\n")
                continue

            match = re.match(r"^(.+\t)([^\t]+)$", row)
            if match is None:
                file.write(row)
                file.write("\n")
                continue

            left = match.group(1)
            right = match.group(2)

            if right == ".":
                file.write(row)
                file.write("\n")
                continue

            fields_list = []
            for v in right.split(";"):
                name, value = v.split("=", 1)
                fields_list.append((name, value))

            attr = OrderedDict(fields_list)

            data = tbldata.find(
                target_name=attr["ID"],
                query_name=attr["Profile_name"],
                query_acc=attr["Profile_acc"],
            )
            if data is None:
                file.write(row)
                file.write("\n")
                continue

            attr["E-value"] = data.full_sequence.e_value
            file.write(left + ";".join(k + "=" + v for k, v in attr.items()))
            file.write("\n")



from typing import Dict, NamedTuple, Optional

__all__ = ["TBLScore", "TBLRow", "TBLData", "TBLIndex"]

TBLScore = NamedTuple("TBLScore", [("e_value", str), ("score", str), ("bias", str)])


TBLRow = NamedTuple(
    "TBLRow",
    [
        ("target_name", str),
        ("target_accession", str),
        ("query_name", str),
        ("query_accession", str),
        ("full_sequence", TBLScore),
        ("best_1_domain", TBLScore),
        ("exp", str),
        ("reg", str),
        ("clu", str),
        ("ov", str),
        ("env", str),
        ("dom", str),
        ("rep", str),
        ("inc", str),
        ("description", str),
    ],
)

TBLIndex = NamedTuple(
    "TBLIndex",
    [
        ("target_name", str),
        ("target_accession", str),
        ("query_name", str),
        ("query_accession", str),
    ],
)


class TBLData:
    def __init__(self, file):
        import csv

        self.rows: Dict[TBLIndex, TBLRow] = {}

        reader = csv.reader(decomment(file), delimiter=" ", skipinitialspace=True)
        for line in reader:
            full_sequence = TBLScore(e_value=line[4], score=line[5], bias=line[6])
            best_1_domain = TBLScore(e_value=line[7], score=line[8], bias=line[9])
            index = TBLIndex(line[0], line[1], line[2], line[3])
            row = TBLRow(
                target_name=line[0],
                target_accession=line[1],
                query_name=line[2],
                query_accession=line[3],
                full_sequence=full_sequence,
                best_1_domain=best_1_domain,
                exp=line[10],
                reg=line[11],
                clu=line[12],
                ov=line[13],
                env=line[14],
                dom=line[15],
                rep=line[16],
                inc=line[17],
                description=line[18],
            )

            if index in row:
                raise RuntimeError("Duplicate tblout index.")
            self.rows[index] = row

    def find(
        self,
        target_name: str = "-",
        target_acc: str = "-",
        query_name: str = "-",
        query_acc: str = "-",
    ) -> Optional[TBLRow]:
        i = TBLIndex(target_name, target_acc, query_name, query_acc)
        return self.rows.get(i, None)


def decomment(rows):
    for row in rows:
        if row.startswith("#"):
            continue
        yield row



import sys
from pathlib import Path

from iseq.file import fetch_file, make_executable, tmp_cwd
from iseq.tblout import TBLData

__all__ = ["HMMSearch"]


_filemap = {
    "hmmsearch_macosx_10_9_x86_64": "6ac1afefae642933b19f4efd87bf869b58c5429e48645ebf0743ef2f03d2fd93",
    "hmmsearch_manylinux2010_x86_64": "74f19eadbb3b43747b569c5e708ed32b1191261c4e332c89c4fd9100fdfaf520",
}


class HMMSearch:
    def __init__(self):

        filename = "hmmsearch"

        if sys.platform == "linux":
            filename = "hmmsearch_manylinux2010_x86_64"
        elif sys.platform == "darwin":
            filename = "hmmsearch_macosx_10_9_x86_64"
        else:
            raise RuntimeError(f"Unsupported platform: {sys.platform}.")

        url_base = "https://iseq-py.s3.amazonaws.com/binhouse"
        prog_path = fetch_file(filename, "bin", url_base, _filemap)

        make_executable(prog_path)
        self._prog_path = prog_path

    def search(self, profile: Path, target: Path) -> TBLData:
        import subprocess

        profile = profile.absolute()
        target = target.absolute()

        with tmp_cwd():
            cmd = [self._prog_path, "--tblout", "tblout", str(profile), str(target)]
            subprocess.check_output(cmd)
            with open("tblout", "r") as file:
                return TBLData(file)



#!/usr/bin/env python

import sys
from setuptools import setup

url = {
    "linux": "https://iseq-py.s3.eu-west-2.amazonaws.com/binhouse/hmmsearch_manylinux2010_x86_64",
    "darwin": "https://iseq-py.s3.eu-west-2.amazonaws.com/binhouse/hmmsearch_macosx_10_9_x86_64",
}

if __name__ == "__main__":
    setup()
